package ru.ulstu.extractor.service.assessment;

import org.springframework.stereotype.Service;
import ru.ulstu.extractor.assessment.model.Assessment;
import ru.ulstu.extractor.rule.model.AssessmentException;
import ru.ulstu.extractor.rule.model.DbRule;
import ru.ulstu.extractor.rule.service.DbRuleService;
import ru.ulstu.extractor.rule.service.FuzzyInferenceService;
import ru.ulstu.extractor.ts.model.TimeSeries;
import ru.ulstu.extractor.ts.service.TimeSeriesService;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

@Service
public class AssessmentService {
    private final DbRuleService ruleService;
    private final TimeSeriesService timeSeriesService;
    private final FuzzyInferenceService fuzzyInferenceService;

    public AssessmentService(DbRuleService ruleService,
                             TimeSeriesService timeSeriesService,
                             FuzzyInferenceService fuzzyInferenceService) {
        this.ruleService = ruleService;
        this.timeSeriesService = timeSeriesService;
        this.fuzzyInferenceService = fuzzyInferenceService;
    }

    public List<Assessment> getAssessmentsByForecastTendencies(Integer branchId) {
        List<TimeSeries> timeSeries = timeSeriesService.getByBranch(branchId);
        List<DbRule> dbRules = ruleService.getList();
        try {
            return getAssessmentsByTimeSeriesTendencies(dbRules, timeSeries);
        } catch (AssessmentException ex) {
            return new ArrayList<>();
        }
    }

    public List<Assessment> getAssessmentsByLastValues(Integer branchId) {
        List<TimeSeries> timeSeries = timeSeriesService.getByBranch(branchId);
        List<DbRule> dbRules = ruleService.getList();
        return getAssessmentsByLastValues(dbRules, timeSeries);
    }

    private List<Assessment> getSingleAssessmentByTimeSeriesTendencies(List<DbRule> dbRules, List<TimeSeries> timeSeries) throws AssessmentException {
        Map<String, Double> variableValues = new HashMap<>();
        timeSeries.forEach(ts -> variableValues.put(ts.getTimeSeriesType().name(),
                timeSeriesService.getLastTimeSeriesTendency(ts)
                        .orElseThrow(() -> new AssessmentException(""))));
        return fuzzyInferenceService.getFuzzyInference(dbRules, variableValues);
    }

    private List<Assessment> getAssessmentsByTimeSeriesTendencies(List<DbRule> dbRules, List<TimeSeries> timeSeries) {
        return dbRules
                .stream()
                .flatMap(dbRule -> {
                    Map<String, Double> variableValues = new HashMap<>();
                    timeSeries
                            .stream()
                            .filter(ts -> ts.getTimeSeriesType() == dbRule.getFirstAntecedent()
                                    || ts.getTimeSeriesType() == dbRule.getSecondAntecedent())
                            .forEach(ts -> variableValues.put(ts.getTimeSeriesType().name(), timeSeriesService
                                    .getLastTimeSeriesTendency(ts)
                                    .orElseThrow(() -> new AssessmentException(""))));
                    return fuzzyInferenceService.getFuzzyInference(List.of(dbRule), variableValues).stream();
                }).collect(Collectors.toList());
    }

    private List<Assessment> getAssessmentsByLastValues(List<DbRule> dbRules, List<TimeSeries> timeSeries) {
        Map<String, Double> variableValues = new HashMap<>();
        timeSeries.forEach(ts -> variableValues.put(ts.getTimeSeriesType().name(), ts.getValues().get(ts.getValues().size() - 1).getValue()));
        return fuzzyInferenceService.getFuzzyInference(dbRules, variableValues);
    }

}
